/* Class: DArray
   This is a templatized dynamic array class that implements a type
   and bounds checked array with an adequate suite of features.  The
   array is implemented as a dynamically allocated array of the type given
   This array is resized by the given factor when automatically
   grown or shrunk.  The array behaves like a normal array, except new methods
   for inserting and removing are added.  The allocated memory and logical 
   lengths of the array are cached.
   */

#ifndef _DARRAY_H_
#define _DARRAY_H_

#include <assert.h>
#include <iostream.h>

template<class Element>
class DArray {
public:
  DArray(int initialLength=0);
  DArray(const DArray& orig);
  DArray<Element>& operator=(const DArray& orig);
  ~DArray() {delete[] array;}

  int length() const {return logLength;}
  void resize(int newLen);

  Element& operator[](int index);
  const Element& operator[](int index) const;

  void insertAt(const Element& el, int index);
  void append(const Element &el);
  void removeAt(int index);
  void remove(const Element& el);

  bool contains(const Element& el);

private:
  Element *array;
  int logLength;
  int allocLength;
  static const increaseFactor = 2;
  static Element *helpAlloc(int size);
};

/* Function: DArray<Element>::DArray
   The default constructor creates an array of the specified length.
   Allocation is handled by a helper function.  The default length is 0.
   */

template<class Element>
DArray<Element>::DArray(int initialLength):
  logLength(initialLength),allocLength(initialLength)
{
  assert(initialLength >= 0);
  array = helpAlloc(initialLength);
}

/* The copy constructor allocates a new, identical array and copies each
   element piecewise.
   */

template<class Element>
DArray<Element>::DArray(const DArray& orig)
{
  logLength = orig.logLength;
  allocLength = orig.allocLength;
  array = helpAlloc(allocLength);
  for (int i = 0; i < logLength; i++)
    array[i] = orig.array[i];
}

/* The assignment operator also copies an entire DArray in a similar manner.
   This operation oly takes place if the target array is distinct from the source.
   */
template<class Element>
DArray<Element>& DArray<Element>::operator=(const DArray& orig)
{
  if (this != &orig) {
    delete[] array;
    logLength = orig.logLength;
    allocLength = orig.allocLength;
    array = helpAlloc(allocLength);
    for (int i = 0; i < logLength; i++)
      array[i] = orig.array[i];
  }
  return *this;
}

/* Function: DArray<Element>::resize
   This function changes the logical length of a DArray to a specified value.
   This function will reallocate the array to make sure that the logical
   length is no smaller than the allocated length divided by the increaseFactor.
   When reallocated, the new elements are copied over piecewise.
   */
template<class Element>
void DArray<Element>::resize(int newLen)
{
  if ((newLen > allocLength) || (newLen < (allocLength / increaseFactor))) {
    Element *newArray;
    newArray = helpAlloc(newLen);
    if (newLen > logLength) {
      for (int i = 0; i < logLength; i++)
	newArray[i] = array[i];
    }
    else {
      for (int i = 0; i < newLen; i++)
	newArray[i] = array[i];
    }
    allocLength = newLen;
    delete[] array;
    array = newArray;
  }
  logLength = newLen;
}

/* Functions: subscript operators
   This function provide a facility for accessing individual elements of the
   DArray.  The parameter is bounds checked.
   */
template<class Element>
Element& DArray<Element>::operator[](int index)
{
  assert((index >= 0) && (index < logLength));
  return array[index];
}

template<class Element>
const Element& DArray<Element>::operator[](int index) const
{
  assert((index >= 0) && (index < logLength));
  return array[index];
}

/* Function: DArray<Element>::insertAt
   This function inserts the specified element (by copying) at the given position
   in the array.  If the array is full, it is scaled by increaseFactor in size.
   Space is made by shifting elements over by copying.
   */

template<class Element>
void DArray<Element>::insertAt(const Element& el, int index)
{
  assert((index >= 0) && (index < logLength));
  if (logLength == allocLength) {
    int savedLength = logLength;
    resize(allocLength * increaseFactor + 1);
    logLength = savedLength;
  }
  for (int i = logLength; i > index; i--)
    array[i] = array[i-1];
  array[index] = el;
  logLength++;
}

// Function: DArray<Element>::append
// This function places the new element at the end of the DArray.
// If the array is full, it is resized as before.

template<class Element>
void DArray<Element>::append(const Element& el)
{
  assert(logLength <= allocLength);
  if (logLength == allocLength) {
    int savedLength = logLength;
    resize(allocLength * increaseFactor + 1);
    logLength = savedLength;
  }
  array[logLength++] = el;
}

// Function: DArray<Element>::removeAt
// This function removes an elemnet and shifs over the others to maintain
// compactness.  Resize is called to change the logical length (and to
// make sure that the allocated length isn't too big.

template<class Element>
void DArray<Element>::removeAt(int index) {
  assert((index >= 0) && (index < logLength));
  for (int i = index; i < logLength - 1; i++)
    array[i] = array[i + 1];
  resize(logLength - 1);
}

// Function: DArray<Element>:remove
// This function removes all matching elements from the DArray.  It depends on
// the operator== for the element.  Resize is called to change the logical
// length.

template<class Element>
void DArray<Element>::remove(const Element& el) {
  int dest = 0;

  for (int src = 0; src < length(); src++) {
    if ((*this)[src] == el) {
      removeAt(src--);
    }
  }
}


// Function: DArray::helpAlloc
// This static helper function allocates a heap array of the specified size
// and asserts it.  It is a fairly simple function.

template<class Element>
Element *DArray<Element>::helpAlloc(int size) {
  Element *newArray;

  if (!size)
    return NULL;
  newArray = new Element[size];
  assert(newArray);
  return(newArray);
}

template<class Element>
bool DArray<Element>::contains(const Element& el) {
  for (int i = 0; i < logLength; i++)
     if (array[i] == el)
        return true;

  return false;
}
#endif
